
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import argparse
import codecs
import logging

from ..util		import log_cfg, log_level, input_secure, ordinal
from .			import recover, recover_bip39, produce_bip39

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( __package__ )


def main( argv=None ):
    ap				= argparse.ArgumentParser(
        description = "Recover and output secret Seed from SLIP-39 or BIP-39 Mnemonics",
        formatter_class = argparse.RawDescriptionHelpFormatter,
        epilog = """\
If you obtain a threshold number of SLIP-39 mnemonics, you can recover the original
secret Seed Entropy, and then re-generate one or more wallets from it.

Enter the mnemonics when prompted and/or via the command line with -m |--mnemonic "...".

The secret Seed Entropy can then be used to generate a new SLIP-39 encoded wallet:

    python3 -m slip39 --secret = "ab04...7f"

SLIP-39 Mnemonics may be encrypted with a passphrase; this is *not* Ledger-compatible, so it rarely
recommended!  Typically, on a Trezor "Model T", you recover using your SLIP-39 Mnemonics, and then
use the "Hidden wallet" feature (passwords entered on the device) to produce alternative sets of
accounts.

BIP-39 Mnemonics can be backed up as SLIP-39 Mnemonics, in two ways:

1) The actual BIP-39 standard 512-bit Seed can be generated by supplying --passphrase, but only at
the cost of 59-word SLIP-39 mnemonics.  This is because the *output* 512-bit BIP-39 Seed must be
stored in SLIP-39 -- not the *input* 128-, 160-, 192-, 224-, or 256-bit entropy used to create the
original BIP-39 mnemonic phrase.

2) The original BIP-39 12- or 24-word, 128- to 256-bit Seed Entropy can be recovered by supplying
--entropy.  This modifies the BIP-39 recovery to return the original BIP-39 Mnemonic Entropy, before
decryption and seed generation.  It has no effect for SLIP-39 recovery.

""" )

    ap.add_argument( '-v', '--verbose', action="count",
                     default=0,
                     help="Display logging information." )
    ap.add_argument( '-q', '--quiet', action="count",
                     default=0,
                     help="Reduce logging output." )
    ap.add_argument( '-m', '--mnemonic', action='append',
                     help="Supply another SLIP-39 (or a BIP-39) mnemonic phrase" )
    ap.add_argument( '-e', '--entropy', action='store_true',
                     default=None,
                     help="Return the BIP-39 Mnemonic Seed Entropy instead of the generated Seed (default: True if --using-bip39 w/o passphrase)" )
    ap.add_argument( '--no-entropy', dest='entropy', action='store_false',
                     help="Return the BIP-39 Mnemonic generated Seed" )
    ap.add_argument( '-b', '--bip39', action='store_true',
                     default=None,
                     help="Recover Entropy and generate 512-bit secret Seed from BIP-39 Mnemonic + passphrase" )
    ap.add_argument( '-u', '--using-bip39', action='store_true',
                     default=None,
                     help="Recover Entropy from SLIP-39, generate 512-bit secret Seed using BIP-39 Mnemonic + passphrase" )
    ap.add_argument( '--binary', action='store_true',
                     default=False,
                     help="Output seed in binary instead of hex" )
    ap.add_argument( '--language',
                     default=None,
                     help="BIP-39 Mnemonic language (default: english)" )
    ap.add_argument( '-p', '--passphrase',
                     default=None,
                     help="Decrypt the SLIP-39 or BIP-39 master secret w/ this passphrase, '-' reads it from stdin (default: None/'')" )
    args			= ap.parse_args( argv )

    log_cfg['level']		= log_level( args.verbose - args.quiet )

    # Set up logging; also, handle the degenerate case where logging has *already* been set up (and
    # basicConfig is a NO-OP), by (also) setting the logging level
    logging.basicConfig( **log_cfg )
    if args.verbose:
        logging.getLogger().setLevel( log_cfg['level'] )
    if args.entropy is None:
        # For SLIP-39, if using_bip39 (unless passphrase supplied), default to recover Seed Entropy
        # (not BIP-39 hashed derivation seed).  This defaults to recovering the BIP-39 encoded seed
        # entropy as Mnemonic.  Use --no-entropy to get it as a derivation Seed instead.
        args.entropy		= not args.bip39 and args.using_bip39 and not args.passphrase
    if args.entropy:
        assert not args.passphrase, "--entropy and --passphrase cannot be used together"
    # Optional passphrase (utf-8 encoded bytes)
    passphrase			= args.passphrase or ""
    if passphrase == '-':
        passphrase		= input_secure( 'Master seed passphrase: ', secret=True )
    elif passphrase:
        log.warning( "It is recommended to not use '-p|--passphrase <password>'; specify '-' to read from input" )
    passphrase			= passphrase.encode( 'UTF-8' )

    # If BIP-39 recovery designated, only a single mnemonic is allowed:
    secret			= None
    algo			= "BIP-39" if args.bip39 else "SLIP-39"
    mnemonics			= args.mnemonic or []
    if args.bip39:
        # Recover from a BIP-39 Mnemonic.  By default, outputs the generated 512-bit wallet Seed,
        # required to derive crypto accounts.  Optionally, with the --entropy option, returns the
        # un-decrypted original 128- to 256-bit Seed Entropy encoded in the BIP-39 Mnemonic phrase
        # (undecrypted; no --passphrase allowed).
        assert 0 <= len( mnemonics ) <= 1, "BIP-39 requires exactly one Mnemonic phrase"
        if not mnemonics:
            try:
                phrase		= input_secure( f"Enter {ordinal(len(mnemonics)+1)} {algo} mnemonic: ", secret=False )
            except KeyboardInterrupt:
                return 0
            mnemonics.append( phrase )
        try:
            secret		= recover_bip39(
                *mnemonics,
                passphrase	= passphrase,
                as_entropy	= args.entropy,
                language	= args.language,
            )
        except KeyboardInterrupt:
            return 0
        except Exception as exc:
            log.error( f"Could not recover {algo} seed with supplied mnemonic: {exc}" )
    else:
        # Collect more mnemonics 'til we can successfully recover the master secret seed
        while secret is None:
            try:
                secret		= recover(
                    mnemonics,
                    passphrase	= passphrase,
                    using_bip39	= args.using_bip39,
                    as_entropy	= args.entropy,
                    language	= args.language,
                )
            except KeyboardInterrupt:
                return 0
            except Exception as exc:
                if mnemonics:
                    log.info( f"Could not recover {algo} seed with {len(mnemonics)} supplied mnemonics: {exc}" )
                try:
                    phrase	= input_secure( f"Enter {ordinal(len(mnemonics)+1)} {algo} mnemonic: ", secret=False )
                except KeyboardInterrupt:
                    return 0
                if ':' in phrase:  # Discard any "<name>: <mnemonic>" name prefix.
                    _,phrase	= phrase.split( ':', 1 )
                mnemonics.append( phrase )
    if secret:
        if args.using_bip39 and args.entropy in (None, True):
            secret		= produce_bip39(
                entropy		= secret,
                language	= args.language,
            )
        else:
            secret		= codecs.encode( secret, 'hex_codec' ).decode( 'ascii' )
            log.info( f"Recovered {algo} secret; To re-generate SLIP-39 wallet, send it to: python3 -m slip39 --secret -" )
            if args.binary:
                secret		= ''.join( f"{int(h,16):0>4b}" for h in secret )
        print( secret )
    return 0 if secret else 1



import json
import random
import math
import logging

from fractions		import Fraction
from pathlib		import Path

import pytest

#import tabulate
from ..			import tabulate

from crypto_licensing.misc import parse_datetime

from ..util		import ordinal, commas
from ..api		import account, accounts
from .artifact		import (
    LineItem, Invoice, InvoiceMetadata, conversions_remaining, conversions_table, Contact,
    produce_invoice, write_invoices,
)

log				= logging.getLogger( "artifact_test" )

SEED_ZOOS			= 'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong'


def test_conversions():
    #print( f"tabulate version: {tabulate.__version__}" )
    c0_tbl = tabulate.tabulate( [[ 1.23],[12345.6789],[.0001234]], floatfmt=",.6g", tablefmt="orgtbl" )
    #print( f"\n{c0_tbl}" )
    assert c0_tbl == """\
|      1.23      |
| 12,345.7       |
|      0.0001234 |"""

    c1				= {
        ('ETH','USD'):	1234.56,
        ('BTC','USD'):	23456.78,
        ('BTC','ETH'): None,
    }
    c1_tbl			= conversions_table( c1, tablefmt='orgtbl' )
    #print( '\n' + c1_tbl )
    assert c1_tbl == """\
| Coin |    in USD | in ETH |
|------+-----------+--------|
| BTC  | 23,456.78 | ?      |
| ETH  |  1,234.56 |        |
| USD  |           |        |"""

    c_simple			= dict( c1 )
    c_simple_i			= 0
    while ( conversions_remaining( c_simple )):
        c_simple_i	       += 1
    assert c_simple_i == 1
    c_simple_tbl		= conversions_table( c_simple, tablefmt='orgtbl' )
    #print( c_simple_tbl )
    assert c_simple_tbl == """\
| Coin |    in USD |     in BTC |    in ETH |
|------+-----------+------------+-----------|
| BTC  | 23,456.78 |            | 19.000113 |
| ETH  |  1,234.56 | 0.05263126 |           |
| USD  |           |            |           |"""

    c_w_doge			= dict( c1, ) | { ('DOGE','BTC'): .00000385, ('DOGE','USD'): None }
    c_w_doge_i			= 0
    while ( conversions_remaining( c_w_doge )):
        c_w_doge_i	       += 1
    assert c_w_doge_i == 2
    assert c_w_doge == {
        ('BTC', 'ETH'): pytest.approx( 19.00011, rel=1/1000 ),
        ('BTC', 'USD'): pytest.approx( 23456.78, rel=1/1000 ),
        ('ETH', 'BTC'): pytest.approx( 0.052631, rel=1/1000 ),
        ('ETH', 'USD'): pytest.approx( 1234.56,  rel=1/1000 ),
        ('USD', 'BTC'): pytest.approx( 4.263159e-05, rel=1/1000 ),
        ('USD', 'ETH'): pytest.approx( 8.100051e-04, rel=1/1000 ),
        ('BTC', 'DOGE'): pytest.approx( 259740.2597, rel=1/1000 ),
        ('DOGE', 'BTC'): pytest.approx( 3.85e-06, rel=1/1000 ),
        ('DOGE', 'ETH'): pytest.approx( 7.31504e-05, rel=1/1000 ),
        ('DOGE', 'USD'): pytest.approx( 0.090308, rel=1/1000 ),
        ('ETH', 'DOGE'): pytest.approx( 13670.45839, rel=1/1000 ),
        ('USD', 'DOGE'): pytest.approx( 11.07314216, rel=1/1000 ),
    }
    c_w_doge_tbl		= conversions_table( c_w_doge, tablefmt='orgtbl' )
    #print( c_w_doge_tbl )
    assert c_w_doge_tbl == """\
| Coin |        in DOGE |         in USD |     in BTC |    in ETH |
|------+----------------+----------------+------------+-----------|
| BTC  | 259,740.26     | 23,456.78      |            | 19.000113 |
| ETH  |  13,670.458    |  1,234.56      | 0.05263126 |           |
| USD  |      11.073142 |                |            |           |
| DOGE |                |      0.0903086 |            |           |"""

    c_w_doge_all		= conversions_table( c_w_doge, greater=False, tablefmt='orgtbl' )
    #print( c_w_doge_all )
    assert c_w_doge_all == """\
| Coin |     in BTC |        in DOGE |      in ETH |         in USD |
|------+------------+----------------+-------------+----------------|
| ETH  | 0.05263126 |  13,670.458    |             |  1,234.56      |
| USD  | 4.263e-05  |      11.073142 |  0.00081001 |                |
| DOGE | 3.85e-06   |                |  7.315e-05  |      0.0903086 |
| BTC  |            | 259,740.26     | 19.000113   | 23,456.78      |"""

    c_bad			= dict( c1, ) | { ('DOGE','USD'): None }
    with pytest.raises( Exception ) as c_bad_exc:
        c_bad_i			= 0
        while ( done := conversions_remaining( c_bad, verify=True )):
            c_bad_i	       += 1
        assert done is None
    assert 'Failed to find ratio(s) for DOGE/USD via BTC/ETH, BTC/USD and ETH/USD' in str( c_bad_exc )
    assert c_bad_i == 1

    assert conversions_remaining( c_bad ) == "Failed to find ratio(s) for DOGE/USD via BTC/ETH, BTC/USD and ETH/USD"

    # Ensure we can handle zero-valued tokens
    c_zero			= dict( c1, ) | {('WEENUS','ETH'): 0, ('ZEENUS','ETH'): 0, ('ZEENUS','WEENUS'): None }
    c_zero_i			= 0
    while ( conversions_remaining( c_zero, verify=True )):
        c_zero_i	       += 1
    assert c_zero_i == 1
    assert c_zero == {
        ('BTC', 'ETH'): pytest.approx( 19.00011, rel=1/1000 ),
        ('BTC', 'USD'): pytest.approx( 23456.78, rel=1/1000 ),
        ('ETH', 'BTC'): pytest.approx( 0.052631, rel=1/1000 ),
        ('ETH', 'USD'): pytest.approx( 1234.56,  rel=1/1000 ),
        ('USD', 'BTC'): pytest.approx( 4.263159e-05, rel=1/1000 ),
        ('USD', 'ETH'): pytest.approx( 8.100051e-04, rel=1/1000 ),
        ('WEENUS', 'BTC'): 0.0,
        ('WEENUS', 'ETH'): 0,
        ('WEENUS', 'USD'): 0.0,
        ('WEENUS', 'ZEENUS'): 1,
        ('ZEENUS', 'BTC'): 0.0,
        ('ZEENUS', 'ETH'): 0,
        ('ZEENUS', 'USD'): 0.0,
        ('ZEENUS', 'WEENUS'): 1,
    }
    c_zero_tbl			= conversions_table( c_zero, tablefmt='orgtbl' )
    #print( c_zero_tbl )
    assert c_zero_tbl == """\
| Coin   |    in USD |     in BTC |    in ETH | in WEENUS | in ZEENUS |
|--------+-----------+------------+-----------+-----------+-----------|
| BTC    | 23,456.78 |            | 19.000113 |           |           |
| ETH    |  1,234.56 | 0.05263126 |           |           |           |
| USD    |           |            |           |           |           |
| WEENUS |      0    | 0          |  0        |           |         1 |
| ZEENUS |      0    | 0          |  0        |         1 |           |"""


line_amounts			= [
    [
        LineItem(
            description	= "Widgets for the Thing",
            units	= 198,
            price	= 2.01,
            tax		= Fraction( 5, 100 ),  # exclusive
            currency	= 'US Dollar',
        ),
        ( 417.879, 19.899, "5% add" ),
    ], [
        LineItem(
            description	= "More Widgets",
            units	= 2500,
            price	= Fraction( 201, 100000 ),
            tax		= Fraction( 5, 100 ),  # exclusive
            currency	= 'ETH',
        ),
        ( 5.27625, .25125, "5% add" ),
    ], [
        LineItem(
            description	= "Something else, very detailed and elaborate to explain in few words, so more elaboration is required",
            units	= 100,
            price	= Fraction( 201, 100000 ),
            tax		= Fraction( 105, 100 ),  # inclusive
            currency	= 'Bitcoin',
        ),
        ( .201, .009571, "5% inc" ),
    ], [
        LineItem(
            description	= "Buy some Holo hosting",
            units	= 12,
            price	= 10101.04030201,
            tax		= Fraction( 5, 100 ),  # inclusive
            currency	= 'HoloToken',
        ),
        ( 127273.107805, 6060.624181, "5% add" ),
    ], [
        LineItem( "Worthless", 12345.6789, currency='ZEENUS' ),
        ( 12345.6789, 0, "no tax"),
    ], [
        LineItem( "Simple", 12345.0001 ),  # currency default None ==> USD
        ( 12345.0001, 0, "no tax" ),
    ], [
        LineItem( "Couchant bogy", units=45, price=9e-5, currency='BTC', tax=1.035 ),
        ( 0.00405, 0.000117961165, "3.5% inc" ),
    ]
]


@pytest.mark.parametrize( "line, amounts", line_amounts )
def test_LineItem( line, amounts ):
    amount,taxes,taxinf	= amounts
    assert line.net() == (
        pytest.approx( amount, abs=10 ** -3 ),
        pytest.approx( taxes,  abs=10 ** -3 ),
        taxinf
    )


vendor				= Contact(
    name	= "Dominion Research & Development Corp.",
    contact	= "Perry Kundert <perry@dominionrnd.com>",
    phone	= "+1-780-970-8148",
    address	= """\
275040 HWY 604
Lacombe, AB  T4L 2N3
CANADA
""",
    billing	= """\
RR#3, Site 1, Box 13
Lacombe, AB  T4L 2N3
CANADA
""",
)

client				= Contact(
    name	= "Awesome, Inc.",
    contact	= "Great Guy <perry+awesome@dominionrnd.com>",
    address	= """\
123 Awesome Ave.
Schenectady, NY  12345
USA
""",
)


def test_tabulate( tmp_path ):
    accounts			= [
        account( SEED_ZOOS, crypto='Ripple' ),
        account( SEED_ZOOS, crypto='Ethereum' ),
        account( SEED_ZOOS, crypto='Bitcoin' ),
    ]
    conversions			= {
        ("BTC","XRP"): 60000,
    }

    total			= Invoice(
        [
            line
            for line,_ in line_amounts
        ],
        accounts	= accounts,
        currencies	= [ "USD", "HOT" ],
        conversions	= dict( conversions ),
    )
    assert total.decimals( 'Bitcoin' ) == 24
    assert total.decimals( 'BTC' ) == 24
    assert total.decimals( 'WBTC' ) == 24
    assert total.decimals( 'ETH' ) == 18
    assert total.decimals( 'HOT' ) == 18
    assert total.decimals( 'USDC' ) == 6

    #print( json.dumps( list( total.pages() ), indent=4, default=str ))

    tables			= list( total.tables() )
    #for t in tables:
    #    print( t )

    # Can't test until we can fake up fixed token values

    tables			= list( total.tables(
        columns=('#', 'Description', 'Qty', 'Currency', 'Coin', 'Price', 'Tax%', 'Taxes', 'Amount', 'Total USDC'),
    ))
    #for t in tables:
    #    print( t )

    # Get the default formatting for line's w/ currencies w/ 0 decimals, 0 value.
    worthless = '\n\n====\n\n'.join(
        f"{table}\n\n{sub}\n\n{tot}"
        for _,table,sub,tot in Invoice(
            [
                line
                for line,_ in line_amounts
                if line.currency in ("ZEENUS", )
            ],
            currencies	= ["HOT", "ETH", "BTC", "USD"],
            accounts	= accounts,
            conversions	= dict( conversions ),
        ).tables(
            tablefmt	= 'presto',
        )
    )
    #print( worthless )
    assert worthless == """\
 Description                                     | Qty |       Price | Tax%   | Taxes | Amount | Coin
-------------------------------------------------+-----+-------------+--------+-------+--------+--------
 Worthless                                       |   1 | 12,345.6789 | no tax |     0 | 12,346 | ZEENUS
-------------------------------------------------+-----+-------------+--------+-------+--------+--------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |     |             |        |     0 |      0 | BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     |             |        |     0 |      0 | ETH
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     |             |        |     0 |      0 | HOT
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     |             |        |     0 |      0 | USDC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     |             |        |     0 |      0 | WBTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     |             |        |     0 |      0 | WETH
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         |     |             |        |     0 |      0 | XRP

 Account                                         | Taxes | Subtotal | Coin | Currency
-------------------------------------------------+-------+----------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |     0 |        0 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |        0 | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |        0 | HOT  | HoloToken
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |        0 | USDC | USD Coin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |        0 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |        0 | WETH | Wrapped Ether
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         |     0 |        0 | XRP  | Ripple

 Account                                         | Taxes | Total | Coin | Currency
-------------------------------------------------+-------+-------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |     0 |     0 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |     0 | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |     0 | HOT  | HoloToken
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |     0 | USDC | USD Coin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |     0 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     0 |     0 | WETH | Wrapped Ether
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         |     0 |     0 | XRP  | Ripple"""  # noqa: E501

    # No conversions of non-0 values; default Invoice currency is USD.  Longest digits should be 2
    # Instead of querying BTC, ETH prices, provide a conversion (so our invoice pricing is static)
    conversions_fixed		= dict( conversions ) | {
        (eth,usd): 1500 for eth in ("ETH","WETH") for usd in ("USDC",)
    } | {
        (btc,eth): 15 for eth in ("ETH","WETH") for btc in ("BTC","WBTC")
    }

    shorter_invoice		= Invoice(
        [
            line
            for line,_ in line_amounts
            if line.currency in ("ZEENUS", None) or line.currency.upper().startswith( 'US' )
        ],
        accounts	= accounts,
        conversions	= conversions_fixed,
    )
    # Include some selected columns
    shorter = '\n\n====\n\n'.join(
        f"{table}\n\n{sub}\n\n{tot}"
        for _,table,sub,tot in shorter_invoice.tables(
            columns	= ('#', 'Description', 'Qty', 'Currency', 'Coin', 'Price', 'Tax%', 'Taxes', 'Amount', 'Total USDC', 'Total BTC'),
            tablefmt	= 'presto',
        )
    )
    #print( shorter )
    assert shorter == """\
 # | Description                                     | Qty | Currency      | Coin   |       Price | Tax%   |      Taxes |          Amount | Total USDC |  Total BTC
---+-------------------------------------------------+-----+---------------+--------+-------------+--------+------------+-----------------+------------+------------
 0 | Widgets for the Thing                           | 198 | US Dollar     | USDC   |      2.01   | 5% add | 19.9       |    417.88       |     417.88 | 0.0185724
 1 | Worthless                                       |   1 | ZEENUS        | ZEENUS | 12,345.6789 | no tax |  0         | 12,346          |     417.88 | 0.0185724
 2 | Simple                                          |   1 | USD           | USDC   | 12,345.0001 | no tax |  0         | 12,345          |  12,762.88 | 0.56723907
---+-------------------------------------------------+-----+---------------+--------+-------------+--------+------------+-----------------+------------+------------
   | BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |     | Bitcoin       | BTC    |             |        |  0.0008844 |      0.56723907 |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Wrapped BTC   | WBTC   |             |        |  0.0008844 |      0.56723907 |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Ethereum      | ETH    |             |        |  0.013266  |      8.508586   |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Wrapped Ether | WETH   |             |        |  0.013266  |      8.508586   |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | USD Coin      | USDC   |             |        | 19.9       | 12,762.88       |            |
   | XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         |     | Ripple        | XRP    |             |        | 53.06      | 34,034.34       |            |

 Account                                         |      Taxes |        Subtotal | Coin | Currency
-------------------------------------------------+------------+-----------------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |  0.0008844 |      0.56723907 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.0008844 |      0.56723907 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.013266  |      8.508586   | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.013266  |      8.508586   | WETH | Wrapped Ether
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E | 19.9       | 12,762.88       | USDC | USD Coin
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         | 53.06      | 34,034.34       | XRP  | Ripple

 Account                                         |      Taxes |           Total | Coin | Currency
-------------------------------------------------+------------+-----------------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |  0.0008844 |      0.56723907 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.0008844 |      0.56723907 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.013266  |      8.508586   | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |  0.013266  |      8.508586   | WETH | Wrapped Ether
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E | 19.9       | 12,762.88       | USDC | USD Coin
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         | 53.06      | 34,034.34       | XRP  | Ripple"""  # noqa: E501

    # Test precision of columns related to Bitcoin; should be 8 decimals throughout, even if WBTC
    # (w/ 6 decimals precision, default 2 decimals display) is used.
    bitcoin_invoice		= Invoice(
        [
            line
            for line,_ in line_amounts
            if line.currency in ("Bitcoin", "BTC")
        ],
        accounts	= accounts,
        conversions	= conversions_fixed,
        # currencies	= None,  # Default: USD
    )

    # Include some selected columns
    bitcoin_info		= list( bitcoin_invoice.tables(
        columns	= (
            '#', 'Description', 'Qty', 'Currency', 'Coin', 'Price', 'Tax%', 'Taxes', 'Amount', 'Total USDC', 'Taxes USDC', 'Total BTC', 'Taxes BTC', 'Total WBTC', 'Taxes WBTC'
        ),
        tablefmt	= 'presto',
    ))
    #for page,table,sub,tot in bitcoin_info:
    #    print( f"Bitcoin Invoice page: {json.dumps( page, indent=4, default=repr )}")

    bitcoin = '\n\n====\n\n'.join(
        f"{table}\n\n{sub}\n\n{tot}"
        for _,table,sub,tot in bitcoin_info
    )
    #print( bitcoin )
    assert bitcoin == """\
 # | Description                                     | Qty | Currency      | Coin |   Price | Tax%     |        Taxes |       Amount | Total USDC | Taxes USDC | Total BTC |  Taxes BTC | Total WBTC | Taxes WBTC
---+-------------------------------------------------+-----+---------------+------+---------+----------+--------------+--------------+------------+------------+-----------+------------+------------+------------
 0 | Something else, very detailed and elaborate to  | 100 | Bitcoin       | BTC  | 0.00201 | 5% inc   |   0.00957143 |      0.201   |   4,522.5  |     215.36 |   0.201   | 0.00957143 |    0.201   | 0.00957143
   | explain in few words, so more elaboration is    |     |               |      |         |          |              |              |            |            |           |            |            |
   | required                                        |     |               |      |         |          |              |              |            |            |           |            |            |
 1 | Couchant bogy                                   |  45 | BTC           | BTC  | 9e-05   | 3.5% inc |   0.00013696 |      0.00405 |   4,613.62 |     218.44 |   0.20505 | 0.00970839 |    0.20505 | 0.00970839
---+-------------------------------------------------+-----+---------------+------+---------+----------+--------------+--------------+------------+------------+-----------+------------+------------+------------
   | BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |     | Bitcoin       | BTC  |         |          |   0.00970839 |      0.20505 |            |            |           |            |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Wrapped BTC   | WBTC |         |          |   0.00970839 |      0.20505 |            |            |           |            |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Ethereum      | ETH  |         |          |   0.145626   |      3.07575 |            |            |           |            |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | Wrapped Ether | WETH |         |          |   0.145626   |      3.07575 |            |            |           |            |            |
   | ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |     | USD Coin      | USDC |         |          | 218.44       |  4,613.62    |            |            |           |            |            |
   | XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         |     | Ripple        | XRP  |         |          | 582.5        | 12,303       |            |            |           |            |            |

 Account                                         |        Taxes |     Subtotal | Coin | Currency
-------------------------------------------------+--------------+--------------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |   0.00970839 |      0.20505 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.00970839 |      0.20505 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.145626   |      3.07575 | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.145626   |      3.07575 | WETH | Wrapped Ether
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E | 218.44       |  4,613.62    | USDC | USD Coin
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         | 582.5        | 12,303       | XRP  | Ripple

 Account                                         |        Taxes |        Total | Coin | Currency
-------------------------------------------------+--------------+--------------+------+---------------
 BTC: bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2 |   0.00970839 |      0.20505 | BTC  | Bitcoin
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.00970839 |      0.20505 | WBTC | Wrapped BTC
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.145626   |      3.07575 | ETH  | Ethereum
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E |   0.145626   |      3.07575 | WETH | Wrapped Ether
 ETH: 0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E | 218.44       |  4,613.62    | USDC | USD Coin
 XRP: rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV         | 582.5        | 12,303       | XRP  | Ripple"""  # noqa: E501

    # Output some invoices
    this			= Path( __file__ ).resolve()
    test			= this.with_suffix( '' )

    date			= parse_datetime( "2021-01-01 00:00:00.1 Canada/Pacific" )

    (paper_format,orientation),pdf,metadata = produce_invoice(
        invoice		= shorter_invoice,
        metadata	= InvoiceMetadata(
            vendor	= vendor,
            client	= client,
            directory	= test,
            date	= date,
            label	= 'Quote',
        ),
        #inv_image	= 'dominionrnd-invoice.png',    # Full page background image
        logo		= 'dominionrnd-logo.png',       # Logo 16/9 in bottom right of header
    )

    #print( f"Invoice metadata: {metadata}" )
    temp		= Path( tmp_path )
    path		= temp / 'invoice-shorter.pdf'
    pdf.output( path )
    #print( f"Invoice saved: {path}" )

    # Finally, generate invoice with all rows, and all conversions from blockchain Oracle (except
    # XRP, for which we do not have an oracle, so must provide an estimate from another source...)
    invoice			= Invoice(
        [
            line
            for line,_ in line_amounts
        ],
        currencies	= ["HOT", "ETH", "BTC", "USD"],
        accounts	= accounts,
        conversions	= {
            ("BTC","XRP"): 60000,
        }
    )
    metadata			= InvoiceMetadata(
        vendor		= vendor,
        client		= client,
        directory	= test,
    )
    (paper_format,orientation),pdf,metadata = produce_invoice(
        invoice		= invoice,
        metadata	= metadata,
        logo		= 'dominionrnd-logo.png',       # Logo 16/9 in bottom right of header
    )

    #print( f"Invoice metadata: {metadata}" )
    temp		= Path( tmp_path )
    path		= temp / 'invoice-complete.pdf'
    pdf.output( path )
    #print( f"Invoice saved: {path}" )


# Generate a sequence of Invoices w/ unique accounts
with open( '/usr/share/dict/words', 'r' ) as words_f:
    words		= list(
        w.strip() for w in words_f.readlines()
    )


line_currencies		= [
    "Bitcoin",
    "BTC",
    "Ethereum",
    "ETH",
    "DAI",
    None,
    "USDC",
    "US Dollar",
    "USD Coin",
    "US Dollar",
    "USDT",
    "Pax Dollar",
    "Ripple",
    "XRP",
    "HoloToken",
    "HOT",
    "Shiba Inu",
    "WEENUS",  # Worthless; will cause Invoice to fail if chosen as one of payment 'currencies'
    "ZEENUS",
]


def line( i, most=None, seen=None ):
    # Sorted from least to most valuable, so c_offset reduces price of items w/ higher-valued currencies
    c_i				= random.randint( 0, len( line_currencies ) - 1 )  # Index eg. [0,12]
    currency			= line_currencies[c_i]
    if most and seen is not None:
        if len( seen ) < most:
            seen.add( currency )
            log.info( f"Invoice line-items contain only currencies {commas( seen, final='and' )}" )
        else:
            currency		= random.choice( tuple( seen ))
            c_i			= line_currencies.index( currency )

    c_offset			= int( round( math.sqrt( c_i )))		# [0,4]

    return LineItem(
        description	= ' '.join(
            random.choice( words )
            for _ in range( random.randint( 1, 5 ))
        ).capitalize(),
        units		= random.randint( 0, 100 ),
        price		= round( random.random(), random.randint( 0, 6 )) * 10 ** ( c_offset - random.randint( 0, 5 )),
        currency	= currency,
        tax		= random.choice( [0, 1] ) + random.randint( 0, 20 ) / 100,
    )


def items( n, most=None, seen=None ):
    if most and seen is None:
        seen			= set()
    for i in range( n ):
        yield line( i, most=most, seen=seen )


conversions			= {
    ("BTC","XRP"): 60000,
}

desired			= 10
paths			= f"../-{desired - 1}"

invoices_to_write		= [
    (
        # - No obvious route between line-item currencies, and invoice currency:
        #
        # Only XRP LineItems, want USDC in 'currencies' payable (as well as 'BTC', 'ETH' and
        # 'XRP' accounts) -- but only {('XRP','BTC'): <ratio>} conversion provided.  Must deduce
        # a common currency for computing remaining conversions ratios, but must include
        # ('BTC',<something>).
        Invoice(
            items( random.randint( 1, 100 ), most=1, seen=set( ('XRP', ) )),
            accounts	= a,
            currencies	= ['USDC'],
            conversions	= dict( conversions )  # | {('BTC','ETH'): None}, # now automatic
        )
        for a in [
            [
                account( SEED_ZOOS, crypto='Ethereum' ),
                account( SEED_ZOOS, crypto='Bitcoin' ),
                account( SEED_ZOOS, crypto='Ripple' ),
            ]
        ]
    ),
    (
        # - Only ETH.  No conversions required.
        Invoice(
            items( random.randint( 1, 100 ), most=1, seen=set( ('ETH', ) )),
            accounts	= a,
            currencies	= ['ETH'],
        )
        for a in [
            [
                account( SEED_ZOOS, crypto='Ethereum' ),
            ]
        ]
    ),
    (
        # - Random combinations of LineItem / Invoice currencies
        #
        # A sequence (generator) of Invoices w/ random LineItems, over a sequence of 'desired'
        # accounts Only 1-3 different cryptos on line-items, 1-3 other invoice currencies each
        # invoice, and 3 Crypto accounts (only 9 cryptocurrencies will typically fit in the
        # conversions table).
        Invoice(
            items( random.randint( 1, 100 ), most=random.randint( 1, 3 )),
            accounts	= a,
            currencies	= list( random.choice( line_currencies ) for _ in range( random.randint( 0, 2 ))) or None,
            conversions	= dict( conversions ),
        )
        for a in zip(
            accounts( SEED_ZOOS, crypto='Ethereum', paths=paths ),
            accounts( SEED_ZOOS, crypto='Bitcoin', paths=paths ),
            accounts( SEED_ZOOS, crypto='Ripple', paths=paths ),
        )
    ),
]


@pytest.mark.parametrize( "invoices", invoices_to_write )
def test_write_invoices( tmp_path, invoices ):
    directory		= Path( tmp_path )
    log.info( f"Invoice Output path: {directory}" )

    metadata		= InvoiceMetadata(
        client		= client,
        vendor		= vendor,
        directory	= directory,
    )

    i				= None
    for i,(name, output) in enumerate( write_invoices(
        ( (invoice,metadata) for invoice in invoices )
    )):
        log.info( f"{ordinal( i )} Invoice {name}: {output}" )
        if isinstance( output, Exception ):
            # Only certain failures are allowed/expected:
            # - Selecting a zero-value Cryptocurrency as a payment currency
            exc_str			= str( output )
            if "Failed to resolve conversion ratios" in exc_str:
                assert 'EENUS' in exc_str, \
                    f"Failed due to selecting zero-valued crypto as payment currency: {exc_str}"




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

#
# HD Wallet Derivation Paths (Standard BIP-44 / Trezor)
#
#     https://wolovim.medium.com/ethereum-201-hd-wallets-11d0c93c87f7
#
# BIP-44 defines the purpose of each depth level:
#    m / purpose’ / coin_type’ / account’ / change / address_index
#
# Use https://iancoleman.io/bip39/ to confirm the derivations
#

# Default group_threshold / required ratios and groups (with varying styles of definition)
GROUP_REQUIRED_RATIO		= 1/2   # default to 1/2 of group members, rounded up
GROUP_THRESHOLD_RATIO		= 1/2   # default to 1/2 of groups, rounded up
GROUPS				= [
    "First1",
    "Second(1/1)",
    "Fam(4)",
    "Frens3/6"
]

THEME				= 'DarkAmber'   # PySimpleGUI Theme
FONTS				= dict(
    sans	= 'helvetica',
    mono	= 'sourcecode',  # UTF-8 monospaced w/ "━": dejavu, inconsolata, noto, overpass, sourcecode
)

#                                  Y      X       Margin
BUSINESS_CARD			= (2,     3+1/2), 1/32  # noqa: E241
CREDIT_CARD			= (2+1/4, 3+3/8), 1/32
INDEX_CARD			= (3,     5),     1/16  # noqa: E241
HALF_LETTER			= (13.5/3,7+3/4), 1/8   # noqa: E241 (actually 2/letter, 3/legal)
THIRD_LETTER			= (13.5/4,7+3/4), 1/8   # noqa: E241 (actually 3/letter, 4/legal)
QUARTER_LETTER			= (10.5/4,7+3/4), 1/8   # noqa: E241 (actually 4/letter, 5/legal)
PHOTO_CARD			= (3+1/2, 5+1/2), 1/16  # prints on 4x6 photo paper w/ 1/4" default outer border

# SLIP-39 Mnemonic Card Sizes
CARD				= 'business'
CARD_SIZES			= dict(
    business	= BUSINESS_CARD,
    credit	= CREDIT_CARD,
    index	= INDEX_CARD,
    half	= HALF_LETTER,
    third	= THIRD_LETTER,
    quarter	= QUARTER_LETTER,
    photo	= PHOTO_CARD,
)

# Paper Wallet Bill Sizes, by default on PAPER format paper
WALLET				= 'quarter'
WALLET_SIZES			= dict(
    half	= HALF_LETTER,
    third	= THIRD_LETTER,
    quarter	= QUARTER_LETTER,
)

PAGE_MARGIN			= 1/4  # Typical printers cannot print within 1/4" of edge

MM_IN				= 25.4
PT_IN				= 72

PAPER				= 'Letter'
PAPER_FORMATS			= dict(
    Letter	= 'Letter',
    Legal	= 'Legal',
    A4		= 'A4',
    Photo	= (int( 4 * MM_IN ), int( 6 * MM_IN )),
)

ORIENTATION			= 'portrait'

# The available GUI controls Layout Options
LAYOUT				= 'Backup'
LAYOUT_OPTIONS			= [
    'Backup',
    'Create',
    'Recover',
    'Pro',
]
LAYOUT_BAK			= 0
LAYOUT_CRE			= 1
LAYOUT_REC			= 2
LAYOUT_PRO			= 3

BITS_DEFAULT			= 128
BITS				= (128, 256, 512)
BITS_BIP39			= BITS + (160, 192, 224)

MNEM_ROWS_COLS			= {
    20:	( 7, 3),		# 128-bit seed
    33:	(11, 3),		# 256-bit seed
    59:	(12, 5),		# 512-bit seed, eg. from BIP-39 (Unsupported on Trezor)
}

# Separators for groups of Mnemonics, and those that indicate the continuation/last line of a
# Mnemonic phrase
MNEM_PREFIX			= {
    20: '{',
    33: '╭╰',
    59: '┌├└',
}
MNEM_LAST			= '╰└{'
MNEM_CONT			= '╭┌├'

BAUDRATE			= 115200

FILENAME_KEYWORDS		= ['name', 'date', 'time', 'crypto', 'path', 'address']
FILENAME_FORMAT			= "{name}-{date}+{time}-{crypto}-{address}.pdf"

# Default Crypto accounts (and optional paths) to generate
CRYPTO_PATHS			= ('ETH', 'BTC')

__d				= "55"
__m				= "88"
__o				= "BB"
__h				= "DD"
__f				= "FF"
COLOR				= [
    # Primary
    f"0x{__o}{__o}{__f}",  # Blue
    f"0x{__o}{__f}{__o}",  # Green
    f"0x{__f}{__o}{__o}",  # Red
    # Secondary
    f"0x{__o}{__f}{__f}",  # Cyan,
    f"0x{__f}{__o}{__f}",  # Magenta
    f"0x{__f}{__f}{__o}",  # Yellow
    # Tertiary
    f"0x{__o}{__h}{__f}",  # Ocean
    f"0x{__o}{__f}{__h}",  # Turquoise
    f"0x{__f}{__o}{__h}",  # Red-Magenta
    f"0x{__h}{__o}{__f}",  # Violet
    f"0x{__h}{__f}{__o}",  # Lime
    f"0x{__f}{__h}{__o}",  # Orange
    # Other
    f"0x{__o}{__h}{__h}",  # Light Cyan
    f"0x{__h}{__o}{__h}",  # Light Magenta
    f"0x{__h}{__h}{__o}",  # Light Yellow
    # Greys
    f"0x{__h}{__h}{__h}",  # Light grey,
    f"0x{__m}{__m}{__m}",  # Medium grey,
    f"0x{__d}{__d}{__d}",  # Dark grey,
]

# We'll default to 30-second intervals for querying Etherscan for Gas, ETH, ERC-20 Pricing info
ETHERSCAN_MEMO_MAXAGE		= 30
ETHERSCAN_MEMO_MAXSIZE		= None
# For token prices, default to 5 minute refreshes
TOKPRICES_MEMO_MAXAGE		= 5*60
TOKPRICES_MEMO_MAXSIZE		= None

SMTP_TO				= "licensing@dominionrnd.com"
SMTP_FROM			= "no-reply@licensing.dominionrnd.com"

# Invoice options.  Presently, only highly-liquid ERC-20 tokens present in Ethereum AMM
# (Automatic Market Maker) systems should be used in invoices, since we use 1Inch's "Off-Chain
# Oracle" smart contract to get current market values.  This prevents us from needing to "trust"
# anyone to obtain current prices for cryptocurrencies -- if you have access to an Ethereum
# blockchain (either locally or via an HTTPS API like Alchemy), then we can securely and reliably
# get current prices.  To avoid conflicts, by convention we upper-case symbols, lower-case full
# names.
INVOICE_FORMAT			= 'totalize'  # 'presto'  # 'orgtbl'
INVOICE_ROWS			= 60  # rows on invoice; each page, about 1/2 that number of line-items
INVOICE_DESCRIPTION_MAX		= 48  # This may seem low; full-precision Prices, 8-dec. Cryptos need room
INVOICE_CURRENCY		= "USD"
INVOICE_PROXIES			= {
    "USD":		"USDC",
    "us dollar":	"USDC",
    "ETH":		"WETH",
    "ethereum":		"WETH",
    "BTC":		"WBTC",
    "bitcoin":		"WBTC",
}

# Invoice times; very explicit about timezones, b/c short zone names are non-deterministic
INVOICE_DUE			= dict( months=1 )      # Default terms: Net 1 month (~30 days)
INVOICE_STRFTIME		= "%c %z %Z"            # "Wed Aug 16 21:30:00 1988 +0000 UTC"




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__          import annotations

import logging
import traceback
import sys

from typing		import Optional
from pathlib		import Path
from enum		import Enum

import crypto_licensing
from crypto_licensing	import licensing
from crypto_licensing.misc import deduce_name


"""
Invoice payments:

    Receiving payments for an invoice requires a round-trip with the vendor, to inform them of the
unique 'seed' assigned to the client.  Unless this data is *guaranteed* to have been delivered to
the vendor, they will never be able to process the received payment, even if the client pays the
generated account.

    Also, to guarantee that the vendor has authorized the licensing

    Therefore, the flow is:

1) Generate an Ed25519 Agent keypair for the client (or load an existing one).

   Useful for signing data available now, with future guarantee that it was indeed the keyholder
   that validated the data at the time.  For example, confirming a Cryptocurrency payment online and
   therefore issuing a "license" signed by the agent.  Later, when offline, the license can be
   loaded, and it can be confirmed that it was considered "valid" by the owner of the Ed25519
   private key (which may now be encrypted or otherwise unavailable), using only the local Agent's
   public key.

   Also useful for generating data (via Diffie-Hellman (D-H) "exchange") that is unique to an
   Client Agent and another party (eg. a Vendor), without communication (so long as each party knows
   the other party's public key).  Thus, with *only* communication of the parties public keys, both
   the Client and the Vendor can agree on some seed data.

2) Generate a unique data for the client.

   Once a level of "security" is agreed upon (see below), the seed data can be produced -- either by
   the Vendor (in a traditional client on-boarding flow, eg. Client # 12345), OR by the Client (by
   generating seed data and allocating it uniquely to the Client (and somehow informing the Vendor,
   so they can receive the payment).

    - Use the Client Agent's private key + Vendor Agent's public key to derive unique D-H secret.
    - Hash with whatever other local date (eg. Machine-ID, account User Name, ...) is required/desired

   With this seed data, we can now proceed to generating payment information.

   - This will be some combination of unique identifying material, associated or signed by Agent privkey

     - Machine ID; unique to one machine; changes on recovery from backup to a new machine
       - It is often a MAC address.  Probably too brittle; want to support recover from backup.

     - User Name; will remain constant across multiple machines, recovery from backups
       - Valid for any user who obtains the Agent keypair, and runs the software under the same user name
       - Valid when software + configuration is copied to any machine w/ the same account user name
       - Probably ok in most cases to allow multiple installations w/ same account user name

     - Just the Agent pubkey
       - Valid for anyone copying the configuration.  Probably too weak.

     - A Diffie-Hellman (D-H) from secrets sk1/2, ie. curve25519(sk1, curve25519(sk2)) == curve25519(sk2, curve25519(sk1))
       - Valid for (derivable by) anyone who holds the Agent keypair, and knows the Vendor public key

3) Send Agent's seed data to Vendor; receive something, eg. "data-signature", ...

   The Client-generated seed data (or the means to derive it) must be communicated to the Vendor, OR
   the Vendor must respond back to the Client with Vendor-generated seed data.

   It is optional to receive response data from the Vendor; usually, a one-way Client -> Vendor data
   flow is adequate.

   However, if a license signed by the Vendor is required (eg. to avoid software modification
   attacks, where the Client subverts the software to generate licenses, eg. in the absence of
   payment, then a response containing the signed license is possible.  However, it doesn't really
   improve the situation -- the client software can simply be subverted to avoid verifying the
   license.

   The normal situation would be to send a DKIM signed email to the Vendor w/ a Subject: line
   containing the seed data, and receive an auto-response eg. containing a verification code.  Since
   the client software originally generated the data (but didn't disclose it to the Agent's human
   author!), the client software itself can generate the DKIM-signed email w/ a Subject: line
   containing the required authentication code, and then await the Agent's author receiving and
   entering the correct code (Ideally, Reed-Solomon error coded to correct substitution/erasure
   errors common to human-transcribed data).

   - Send Client-generated data to the Vendor

     - Write it to the blockchain, eg. w/ MultiPayoutERC20.forwarder/forwarder_allocate
       - A 256-bit arbitrary data is associated w/ a _salt

     - Send DKIM-signed Email via ..communications.send_message w/ a Subject containing the data
       - The response' Subject may contain Vendor-generated data, eg. signature, challenge/response code

   - Receive Vendor-generated data

     - Via a standard Vendor's Client on-boarding flow
       - Client enter address, Vendor assigns and emails seed data

     - In response to a spontaneous DKIM-signed Email from the Client
       - Client sends email, optionally w/ some data (eg. public key), Vendor assigns/email data

   - Obtain vendor pubkey via DKIM from the Vendor's DNS
       - The  "data-signature" + "data" can be verified w/ the Vendor's pubkey

   - Use the "data-signature" 1st 256 bits (32 bytes) as the seed, store 2nd 32 bytes as its data?

     - This stores the entire signature in the MultiPayoutERC20 contract, proving that...

   - Verified "data" + "data-signature" + observation of payment (on blockchain)

4) Generate Client-unique Cryptocurrency address(es) and Invoice

   Now, the Client's software has sufficient seed data to generate the client-unique addresses.
   These could simply be generated from an xpub... key, at a HD Wallet derivation path derived from
   the seed data.  Or, they could use the seed data as the "salt" in an Ethereum CREATE2 contract
   address calculation, to generate single-purpose forwarding addresses.

   Either way, once the Ethereum, Bitcoin, ... addresses are generated, the Invoice can be generated
   for the desired functionality line-items, for whatever Fiat or Crypto price, payable in various
   Cryptocurrencies to the designated addresses.


5) Detect payment to the addresses

   Consult the blockchain (directly via a local node, or indirectly using one of the available
   public blockchain transaction services such as Alchemy or Infura, or monitoring services such as
   Etherscan, etc.  See .ethereum.* for various APIs for ETH and EEC-20 Tokens.

"""
log				= logging.getLogger( "payments" )


class Process( Enum ):
    FAILED	= 0		# Failed to process; val is the failure text
    PROMPT	= 1		# Some user input required; val is the prompt text
    KEYPAIR	= 2		# A Keypair, but no associated License(s); val is (Keypair,None)
    LICENSE	= 3		# Some Keypair/License; val is (Keypair,License)
    GRANTS	= 4		# Some License grants; val is the Grants


def reload(
    author: licensing.Agent,			# Details of the author's product we're licensing
    client: Optional[licensing.Agent] = None,   # ..and the intended specific client Agent
    username: Optional[str]	= None,		# The credentials for our client Agent's Keypair
    password: Optional[str]	= None,
    basename			= None,
    registering			= None,
    acquiring			= None,
    **kwds
):
    """Attempts to reload any existing client Agent Keypairs, and any License(s) authored to that
    client, and yields any (Process.GRANT, <Grants>) found.  If Keypair(s) are found, they will be
    yield (Process.KEYPAIR, <Keypair>).

    During the process, may yield 0 or more (Process.PROMPT, <str>) requesting input from the
    caller, which may be supplied via <generator>.send( ... ).

    If an error occurs (such as no Grants found), a (Process.ERROR, <str>) will be produced.  This
    will often be

    """
    # If no username/password provided, we'll loop once w/ None, then request input for subsequent authorizations
    cl_username			= '-' if username is None else username
    if cl_username == '-':
        username		= None
    cl_password			= '-' if password is None else password
    if cl_password == '-':
        password	        = None
    userpass_input		= cl_username == '-' or cl_password == '-'

    try:
        authorizations		= licensing.authorized(
            author	= author,
            client	= client,
            username	= username,
            password	= password,
            basename	= basename,
            registering	= registering,          # Issue an Agent ID if none found?
            acquiring	= acquiring,		# Acquire a License if none found?
            **kwds
        )

        loaded			= []
        try:
            # Iterate using authorizations.send(...) instead of next; if nothing to send, leave 'send' None
            credentials		= None
            while True:
                key,lic		= authorizations.send( credentials )
                log.detail( f"Reloading key: {key}, lic: {lic}" )
                credentials	= None
                if key is None or lic is None:
                    if key is not None:
                        # Found a Keypair (but no License); yield it
                        yield Process.KEYPAIR, key
                        continue

                    # Neither Keypair nor License; authorized has indicated it is out of options and
                    # is about to give up: unless we provide new credentials.
                    what		= "No License found for Agent ID {}".format(
                        licensing.into_b64( key.vk )) if key else "(No Agent ID Keypair found)"
                    if userpass_input:
                        what       += "; enter credentials"
                        if password != "-":
                            what   += " (leave blank to register w/ {}: {}".format(
                                username or "(no username)", '*' * len( password or '' ) or "(no password)" )
                    log.warning( what )
                    if not userpass_input:
                        continue
                    # No Agent ID/License loaded; username/password may be incorrect.  If none provided,
                    # then authorization may go on to register w/ the last-entered username/password.
                    # Either username or password may be updated, if desired.  Usually, credential input
                    # forces you to re-enter something you know to be correct; this loop does not.
                    # Failing to enter both credentials indicates satisfaction -- goes on to register a
                    # new Agent ID w/ credentials, if so.
                    userpass_updated = False
                    if cl_username == '-':
                        log.warning( "Asking for username..." )
                        username_update	= (
                            yield Process.PROMPT, "Enter {} username (leave empty for no change): ".format(
                                deduce_name( basename=basename, filename=kwds.get( 'filename' ), package=kwds.get( 'package' ),
                                             default=client and client.servicekey or "the" ))
                        )
                        userpass_updated |= bool( username_update )
                        if username_update:
                            username	= username_update
                    if cl_password == '-':
                        log.warning( "Asking for password..." )
                        password_update	= (
                            yield Process.PROMPT, "Enter {} password (leave empty for no change): ".format(
                                deduce_name( basename=basename, filename=kwds.get( 'filename' ), package=kwds.get( 'package' ),
                                             default=client and client.servicekey or "the" ))
                        )
                        userpass_updated |= bool( password_update )
                        if password_update:
                            password	= password_update
                    if userpass_updated:
                        log.detail( "Supplying new credentials for {}: {}".format(
                            username or "(no username)", '*' * len( password or '' ) or "(no password)" ))
                        credentials	= (username,password)
                        continue
                    log.detail( "No new credential(s) for {}: {}{}".format(
                        username or "(no username)", '*' * len( password or '' ) or "(no password)",
                        " (attempting to register new Agent ID)" if registering else " (authorization failed)" ))
                    # No Keypair (or perhaps a Keypair, but no License) found; no new credentials
                else:
                    # Found a Keypair and License; remember/yield it
                    loaded.append( (key,lic) )
                    yield Process.LICENSE, (key,lic)

        except StopIteration:
            log.detail( f"Completed licensing.authorized w/ {len(loaded)} Keypair/Licenses found" )

        assert loaded, \
            "Unable to find {}'s product {!r} service key {!r}; no Licenses found".format(
                author.name, author.product, author.servicekey )

        # Collect up all the License grants; there may be more than one, if the user has purchased
        # multiple Licenses at different times.  Ensures we only include a specific License once.
        grants			= licensing.Grant()
        once			= set()
        for key,lic in loaded:
            grants_lic		= lic.grants( once=once )
            log.normal( "Located Agent Ed25519 Keypair {pubkey} w/ {product} License (from {_from}){extra}".format(
                pubkey	= licensing.into_b64( key.vk ),
                product	= lic and lic.license.author.product or "End-User",
                _from	= "from {}".format( lic._from ) if lic._from else "locally issued",
                extra	= (( " w/ grants: {}".format( grants_lic ) if log.isEnabledFor( logging.DEBUG ) else "" )
                           + ( " merging w/ grants: {}".format( grants ) if log.isEnabledFor( logging.TRACE ) else "" )),
            ))
            grants	       |= grants_lic

        # And, finally: ascertain whether we've collected a Grant to run the Author's product
        assert author.servicekey in grants, \
            "Unable to find {}'s product {!r} service key {!r} in License Grants {}".format(
                author.name, author.product, author.servicekey, grants )

    except Exception as exc:
        log.error( "Failed loading Agent Keypair and/or License: {exc}".format(
            exc=''.join( traceback.format_exception( *sys.exc_info() )) if log.isEnabledFor( logging.TRACE ) else exc ))
        with open( Path( crypto_licensing.__file__ ).resolve().parent / 'licensing' / 'static' / 'txt' / 'CL-KEYPAIR-MISSING.txt', 'r' ) as f:
            error		= f.read().format(
                DISTRIBUTION	= deduce_name( basename=basename, filename=kwds.get( 'filename' ), package=kwds.get( 'package' ), default=client and client.servicekey or "" ),
                KEYPATTERN	= licensing.KEYPATTERN,
                LICENSE_OPTION	= '--license',
            )
        yield Process.FAILED, error

    else:
        yield Process.GRANTS, grants


def process(
    author: Optional[licensing.Agent]	= None,			# Details of the author's product we're licensing
    client: Optional[licensing.Agent]	= None,			# ..and the intended specific client Agent
    username: Optional[str]	= None,			# The credentials for our client Agent's Keypair
    password: Optional[str]	= None,
    basename		= None,
    registering			= None,
    acquiring			= None,
    **kwds
):
    """Load a License, or Process payment for and obtain a License, for certain client capabilities.

    Yields a sequence of prompts for any client information required, and yields the various
    credentials loaded/obtained, and finally (if successful) yields the License.

    Once the caller obtains the License, it should be verified.

    On the simplest path, reloads pre-existing Agent keypair and License files and yields them in
    order.  If the License(s) contain the requested capabilities Grants, then the process is
    complete.

    The Licenses may be associated with a Machine-ID, or have a limited lifetime; therefore, even if
    the Licenses may be loaded, they may not be valid (ie. if they've been loaded onto a new machine
    from backup, or have expired).

    If a License expires, it should contain the original Grants, and their costs (at the time the
    License was issued).  If a new License must be issued, the Grants/costs from the existing
    License must be respected in the new License, so that existing (historical) Cryptocurrency
    payments satisfy them.  Therefore, each Grant's LineItem data must be deducable from the
    existing License(s).  We must obtain the existing Licenses, then check them for validity.


    If certain Grants are insufficient (eg. if we can find Licenses containing a capability 5, and
    the caller requires 10, or if a certain capability is completely absent in the loaded Grants),
    then the remaining Grant must be acquired.

    If required Grants remain, then we attempt to process a variety of approaches to obtain one:

    1) The cost of the Grant is computed, and Invoice LineItems are produced enumerating them

    2) An Invoice PDF if produced, computing prices in various Cryptocurrencies

    3) The Invoice is approved via a PROMPT; when returned

        - if not approved, the process fails

    4) Payment is awaited via another PROMPT; when returned

       - the designated Cryptocurrency's blockchain is queried for payment
       - if no cryptocurrency designated, the process fails
       - if payment not (yet) received, issue an ERROR and loop awaiting payment

    5) When payment received, generate the appropriate License

       - Obtain them from a crypto-licensing server, via its signed API.  These will be signed
         by the Author's Agent,  and will be verifiable via DKIM (check=True).

       - Obtain them locally.  These will be signed by the local Client's Agent, and therefore will
         NOT be verifiable (check=False).  They will, however, provide proof that the local Client
         Agent produced them.

         - Of course, anyone versed in the state of the art can trivially take their local Agent's
           Ed25519 Keypair, and sign anything they wish - including such a License.  However, we are
           not protecting ourselves from determined thieves; we're supporting clients who wish to
           pay for and legitimately use our software.

    """
    if registering is None:
        # If we're provided w/ a client Agent (w/ a keypair or pubkey), we won't be registering one
        registering		= True if client is None else False
    if acquiring is None:
        acquiring		= True

    # We're looking for a License authored by an author Agent (eg. Dominion R&D Corp.), for a
    # certain product or service.  This might be encapsulated as one of the dependencies of the
    # License we find (eg. if Dominion issues a License to some company, which includes the ability
    # to run a crypto-licensing server), but we'll validate that the Grant was issued by Dominion
    # R&D Corp.  For example, Dominion issues a License to awesome-inc.com to sub-License
    # crypto-licensing servers.  They, in turn, issue a License to Лайка.ru to run a
    # crypto-licensing server, and help them set it up.  When
    # http://crypto-licensing.xn--80aa0aec.ru/ (crypto-licensing.Лайка.ru) issues a License for
    # their software, part of the fees are paid to awesome-inc.com and some to dominionrnd.com.  The
    # final software installation uses crypto-licensing's authorized() function, which checks that
    # each successive License's dependencies are correctly signed, and carries the original
    # 'crypto-licensing' Grant through to the recipient.
    dominion			= licensing.Agent(
        name	= licensing.COMPANY,
        domain	= licensing.DOMAIN,
        product	= "SLIP-39",
        pubkey	= licensing.PUBKEY,
    )

    reloader			= reload(
        author		= author or dominion,
        client		= client,
        username    	= username,
        password	= password,
        basename	= basename,
        registering	= registering,
        acquiring	= acquiring,
        **kwds
    )

    keypairs			= []
    grants			= None
    credentials			= None
    try:
        while True:
            key,val		= reloader.send( credentials )
            if key is Process.PROMPT:
                # User information is required; yield the prompt, send the returned data into reloader
                log.warning( f"Prompt: {val}" )
            # Some other key,val not returning input
            if key is Process.KEYPAIR:
                log.warning( f"Keypair: {val}" )
                keypairs.append( val )
            if key is Process.LICENSE:
                log.warning( f"License: {val[1]}" )
                keypairs.append( val[0] )
            if key is Process.GRANTS:
                grants		= val
            credentials		= ( yield (key,val) )
    except StopIteration:
        pass

    if grants:
        # TODO: check grants
        pass

    # No/insufficient license grants.  Engage the process of paying for and obtaining a License.



import json
import logging
import os
import time

import pytest

from web3		import Web3		# noqa F401

from ..util		import timer, ordinal
from .ethereum		import Etherscan, Chain, gasoracle, alchemy_url, tokenprices, tokenratio

log				= logging.getLogger( 'ethereum_test' )


def test_etherscan_ethprice():
    """Check that defaults for ETH pricing work, and that obtaining data from https://etherscan.io works
    even without an ETHERSCAN_API_TOKEN (subject to default 1/5s rate limit.

    """
    ETH				= Etherscan( 'Nothing' )  # Specify Nothing for Ethereum back-end (uses defaults)

    log.warning( "ETH: USD${}".format( ETH.ETH_USD ))
    log.warning( "Gas: USD${:7.2f} / 100,000 gas".format( 100000 * ETH.GAS_GWEI / ETH.ETH_GWEI * ETH.ETH_USD ))

    gas				= 21000
    spend			= 1.50
    gas_pricing			= ETH.maxPriorityFeePerGas( spend=spend, gas=gas )
    log.warning( f"{ETH.chain}: Gas Pricing EIP-1559 for max $1.50 per 21,000 Gas transaction: {json.dumps( gas_pricing )}" )
    assert gas_pricing['maxFeePerGas'] == pytest.approx( 71428571428.57143 )
    assert gas_pricing['maxPriorityFeePerGas'] == pytest.approx( 2000000000.0 )

    # Now try the real Etherscan API.  Will fail if we don't have network access to
    # https://etherscan.io.  If we don't have an API token, observe that the rate limit is applied.
    # Otherwise, if we do have an API token, warn if we observe a rate limit, or fail to get data.
    ETH.chain			= Chain.Ethereum
    if ETH.ETH_USD == ETH.ETH_USD_DEFAULT or not ETH.TIMESTAMP:
        log.warning( f"Should have successfully obtained Ethereum price from Etherscan, regardless of ETHERSCAN_API_TOKEN: {os.getenv( 'ETHERSCAN_API_TOKEN' )}" )
        return
    # We've successfully reached Etherscan once!  Should eventually succeed again.
    beg				= timer()
    while timer() - beg < 10 and not ETH.UPDATED:  # will loop 'til time exhausted or retry succeeeds
        gas_pricing		= ETH.maxPriorityFeePerGas( spend=1.50, gas=21000 )
        now			= timer()
        log.warning( f"{ETH.chain}: After {now - beg:.2f}s ({now-gasoracle.lst:.2f}s since last try): Gas Pricing EIP-1559 for max $1.50 per 21,000 Gas transaction: {json.dumps( gas_pricing )}" )
        if ETH.UPDATED:  # Ignore, if we don't have communications w/ the Ethereum chain data source
            assert gas_pricing['maxFeePerGas'] * gas / ETH.ETH_WEI * ETH.ETH_USD == pytest.approx( spend )
        time.sleep( 1 )
    now				= timer()
    assert ETH.UPDATED, \
        f"Failed to obtain Gas Oracle after {now-gasoracle.lst:.2f}s since {ordinal(gasoracle.cnt)} try"


def test_1inch_offchainoracle():
    """Consult the 1inch offChainOracle contract to query several continuous liquidity pools for Crypto prices.

    """

    # Connect to the real Ethereum blockchain
    w3				= Web3( Web3.WebsocketProvider( alchemy_url( Chain.Ethereum )))

    # See: https://github.com/1inch/spot-price-aggregator/blob/master/examples/multiple-prices.js
    # multicall_abi		= [{"inputs":[{"components":[{"internalType":"address","name":"to","type":"address"},{"internalType":"bytes","name":"data","type":"bytes"}],"internalType":"struct MultiCall.Call[]","name":"calls","type":"tuple[]"}],"name":"multicall","outputs":[{"internalType":"bytes[]","name":"results","type":"bytes[]"},{"internalType":"bool[]","name":"success","type":"bool[]"}],"stateMutability":"view","type":"function"}]  # noqa: E501
    # multicall_contract		= w3.eth.contract( multicall_abi, '0x07D91f5fb9Bf7798734C3f606dB065549F6893bb' )

    offchainoracle_abi		= [{"inputs":[{"internalType":"contract MultiWrapper","name":"_multiWrapper","type":"address"},{"internalType":"contract IOracle[]","name":"existingOracles","type":"address[]"},{"internalType":"enum OffchainOracle.OracleType[]","name":"oracleTypes","type":"uint8[]"},{"internalType":"contract IERC20[]","name":"existingConnectors","type":"address[]"},{"internalType":"contract IERC20","name":"wBase","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"ConnectorAdded","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"ConnectorRemoved","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract MultiWrapper","name":"multiWrapper","type":"address"}],"name":"MultiWrapperUpdated","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IOracle","name":"oracle","type":"address"},{"indexed":False,"internalType":"enum OffchainOracle.OracleType","name":"oracleType","type":"uint8"}],"name":"OracleAdded","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IOracle","name":"oracle","type":"address"},{"indexed":False,"internalType":"enum OffchainOracle.OracleType","name":"oracleType","type":"uint8"}],"name":"OracleRemoved","type":"event"},{"anonymous":False,"inputs":[{"indexed":True,"internalType":"address","name":"previousOwner","type":"address"},{"indexed":True,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnershipTransferred","type":"event"},{"inputs":[{"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"addConnector","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract IOracle","name":"oracle","type":"address"},{"internalType":"enum OffchainOracle.OracleType","name":"oracleKind","type":"uint8"}],"name":"addOracle","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"connectors","outputs":[{"internalType":"contract IERC20[]","name":"allConnectors","type":"address[]"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"srcToken","type":"address"},{"internalType":"contract IERC20","name":"dstToken","type":"address"},{"internalType":"bool","name":"useWrappers","type":"bool"}],"name":"getRate","outputs":[{"internalType":"uint256","name":"weightedRate","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"srcToken","type":"address"},{"internalType":"bool","name":"useSrcWrappers","type":"bool"}],"name":"getRateToEth","outputs":[{"internalType":"uint256","name":"weightedRate","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"multiWrapper","outputs":[{"internalType":"contract MultiWrapper","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"oracles","outputs":[{"internalType":"contract IOracle[]","name":"allOracles","type":"address[]"},{"internalType":"enum OffchainOracle.OracleType[]","name":"oracleTypes","type":"uint8[]"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"removeConnector","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract IOracle","name":"oracle","type":"address"},{"internalType":"enum OffchainOracle.OracleType","name":"oracleKind","type":"uint8"}],"name":"removeOracle","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"renounceOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract MultiWrapper","name":"_multiWrapper","type":"address"}],"name":"setMultiWrapper","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"}]  # noqa: E501

    offchainoracle_contract	= w3.eth.contract( address='0x07D91f5fb9Bf7798734C3f606dB065549F6893bb',  abi=offchainoracle_abi)

    token			= dict(
        address		= '0xdAC17F958D2ee523a2206206994597C13D831ec7',
        decimals	= 6,
    )

    token_price			= offchainoracle_contract.functions.getRateToEth( token['address'], True ).call()
    log.info( f"ETH/USDT: {token_price}" )

    eth_usdt			= token_price / (10 ** ( 18 + 18 - token['decimals'] ))
    log.info( f"ETH/USDT: {eth_usdt}" )
    log.info( f"USDT/ETH: {1/eth_usdt}" )


HOT			= "0x6c6EE5e31d828De241282B9606C8e98Ea48526E2"
USDT			= "0xdAC17F958D2ee523a2206206994597C13D831ec7"
USDC			= "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"
WBTC			= "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
WETH			= "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"


def test_tokenprice():
    """Get prices vs. the default base (ETH), and a low-value token.  Should retain precision,
    due to underlying 10^18 fixed-point offset."""
    for base in (None, HOT):
        prices_eth			= list( tokenprices( USDC, USDT, WBTC, WETH, base=base ))
        log.info( f"Token prices: " + json.dumps( prices_eth, indent=4, default=str ))

        for t,b,f in prices_eth:
            log.info( f"{t.symbol:>6}/{b.symbol:<6}: {float( f ):13.4f} =~= {str(f)}" )

        # WETH in terms of ETH should be close to 1.0 (ignore for any other base currency)
        assert base is not None or float( prices_eth[-1][2] ) == pytest.approx( 1.0 ), \
            "WETH/ETH is diverging from unity"

        # WBTC, WETH vs. ETH should be multiples of USDC vs. ETH
        assert float( prices_eth[-1][2] ) / float( prices_eth[0][2] ) > 1000, \
            "WETH/USDC is inverted?"
        assert float( prices_eth[-2][2] ) / float( prices_eth[0][2] ) > 100, \
            "WBTC/USDC is inverted?"


def test_tokenratio():
    # Now, get the token price ratio of two tokens (both cached, relative to default ETH)
    WBTC_USDC		= tokenratio( WBTC, USDC )
    log.info( f"{WBTC_USDC[0].symbol:>6}/{WBTC_USDC[1].symbol:<6}: {float( WBTC_USDC[2] ):13.4f} =~= {WBTC_USDC[2]}" )
    assert WBTC_USDC[2] > 10000, \
        "WBTC has crashed vs. USDC?"

    HOT_USDC		= tokenratio( "HOT", "USDC" )
    log.info( f"{HOT_USDC[0].symbol:>6}/{HOT_USDC[1].symbol:<6}: {float( HOT_USDC[2] ):13.4f} =~= {HOT_USDC[2]}" )
    assert HOT_USDC[2] < .10, \
        "HOT has exploded vs. USDC?"



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.1.dev600+g6605467.d20230217'
__version_tuple__ = version_tuple = (0, 1, 'dev600', 'g6605467.d20230217')




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import logging
import re
import subprocess
import sys

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( __package__ )


def printers_available():
    """
    On macOS, find any available printers that appear to be available:

        printer Canon_G6000_series is idle.  enabled since Thu  3 Mar 09:24:56 2022
                ...
                Description: Canon G6000
                Alerts: none
                ...
        printer Marg_s_Brother_HL_L2360D is idle.  enabled since Wed 15 Sep 12:19:48 2021
                ...
                Description: Marg's Brother HL-L2360D
                Alerts: offline-report
                ...

    Yields a sequence of their names: (<system>,<long>), ...
    """
    if sys.platform == 'darwin':
        command		= [ '/usr/bin/lpstat', '-lt' ]
        command_input	= None
    else:
        raise NotImplementedError( f"Printing not supported on platform {sys.platform}" )

    subproc			= subprocess.run(
        command,
        input		= command_input,
        capture_output	= True,
        encoding	= 'UTF-8'
    )
    assert subproc.returncode == 0 and subproc.stdout, \
        f"{' '.join( command )!r} command failed, or no output returned"

    if sys.platform == 'darwin':
        system,human,ok		= None,None,None
        for li in subproc.stdout.split( '\n' ):
            printer		= re.match( r"^printer\s+([^\s]+)", li )
            if printer:
                system,human,ok	= printer.group(1).strip(),None,None
                continue
            descr		= re.match( r"^\s+Description:(.*)", li )
            if descr:
                human		= descr.group(1).strip()
                continue
            alerts		= re.match( r"^\s+Alerts:(.*)", li )
            if alerts:
                ok		= 'offline' not in alerts.group(1)
                log.info( f"Printer: {human}: {alerts.group(1)}" )
            if ok:
                yield system,human
                system,human,ok	= None,None,None


def printer_output(
    binary,			# Raw data for printer
    printer		= None,
    orientation		= None,
    paper_format	= None,
):
    """Output raw binary data directly to the printer, eg.

        ... | lpr -P "Canon_G6000_series" -o media=Letter -o sides=one-sided -o job-sheets=secret

    """
    if sys.platform == 'darwin':
        command			= [ '/usr/bin/lpr', '-o', 'sides=one-sided' ]
        command_input		= binary

        # Find the desired printer's system name; otherwise use default printer
        printer_system		= None
        if printer:
            printer_list	= list( printers_available() )
            for system,human in printer_list:
                if human.lower() == printer.lower() or system.lower() == printer.lower():
                    printer_system = system
            assert printer_system, \
                f"Couldn't locate printer matching {printer!r}, in {', '.join( h for s,h in printer_list )}"
        if printer_system:
            command	       += [ '-P', printer_system ]
        if paper_format:
            command    	       += [ '-o', f"media={paper_format.capitalize()}" ]
        if orientation:
            # -o orientation-requested=N   Specify portrait (3) or landscape (4) orientation
            N		= { 'p': 3, 'l': 4 }[orientation.lower()[0]]
            command	       += [ '-o', f"orientation-requested={N}" ]

    log.info( f"Printing via: {' '.join( command )}" )
    subproc			= subprocess.run(
        command,
        input		= command_input,
        capture_output	= True,
    )
    assert subproc.returncode == 0, \
        f"{' '.join( command )!r} command failed"



import pytest
import logging
import subprocess

from pathlib		import Path


from crypto_licensing	import licensing


from ..util		import ordinal
from .payments		import Process, reload

log				= logging.getLogger( "payments_test" )


def test_sending():
    """Test the transmission and receipt of data via iterator.send(...)"""
    def generator( count ):
        for n in range( count ):
            vo			= ordinal( n )
            #print( f"gen. yield: {n, vo}" )
            v,			= ( yield n, vo )
            #print( f"gen. recvd: {v}" )
            assert v == vo

    gi				= generator( 3 )
    try:
        credentials		= None
        while True:
            #print( f"iter send:  {credentials}" )
            n,vo		= gi.send( credentials )
            credentials		= None
            #print( f"iter next:  {n,vo}" )
            credentials		= (ordinal( n ),)
    except StopIteration:
        log.info( f"done w/ n == {n}, s == {vo}" )


def test_grants( tmp_path ):
    """Test self-signed Licensing.  This allows a Vendor to ship a "minimal" set of capabilities, for
    which the Client can pay, and issue a self-signed License to use.

    Of course, since it is self-signed, any Thief knowledgeable in the state of the art can issue
    themselves a License *without* paying!  So, don't use this method to ship highly valuable
    capabilities.

    Of course, any Thief knowledgeable in the state of the art can simply *remove* your License
    checking, and run the software anyway...

    """

    test			= Path( __file__ ).resolve()		# Our payments_test.py file
    here			= Path( tmp_path ).resolve()		# Our testing directory.

    name_ss			= "self-signed"
    base_ss			= here / name_ss
    seed_ss			= licensing.into_bytes( "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" )
    user_ss			= "sales@self-signed.com"
    pswd_ss			= "password"
    keyp_ss			= licensing.registered(
        basename	= str( base_ss ),
        seed		= seed_ss,
        username	= user_ss,
        password	= pswd_ss,
    )
    assert keyp_ss['vk'] == "5zTqbCtiV95yNV5HKqBaTEh+a0Y8Ap7TBt8vAbVja1g="

    author			= licensing.Agent(
        name		= licensing.COMPANY,
        domain		= "self-signed.com",
        product		= "Self Signed",
        keypair		= keyp_ss.into_keypair( username=user_ss, password=pswd_ss ),
    )
    log.info( f"Author: {author}" )

    # Now create the License suitable for self-signing; no client (so any client can sign it)
    lic				= licensing.license(
        basename	= str( base_ss ),
        author		= author,
        grant		= {'self-signed': { 'some-capability': 10 }},
        confirm		= False,  # Not a real domain...
    )
    log.info( f"Self-Signable License: {lic}" )

    # We've got a self-signed.crypto-license, signed by the Vendor, with no specific Client
    # specified.  It's good for anyone to use.  Lets make a link to it, under the name that our
    # client will use.

    name_cl			= "client-user"
    base_cl			= here / name_cl

    # os.symlink( name_ss + '.crypto-license', base_cl.with_suffix( '.crypto-license' ))

    # Get a Client Agent Keypair, to use to self-sign the License
    seed_cl			= licensing.into_bytes( "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb" )
    user_cl			= "end-user@client.com"
    pswd_cl			= "password"
    keyp_cl			= licensing.registered(
        basename	= str( base_cl ),
        seed		= seed_cl,
        username	= user_cl,
        password	= pswd_cl,
    )
    assert keyp_cl['vk'] == "fVnFYj3UCnSqTVoyrGRdOz+V2urkwiviVHbdakhvc4I="
    keyp_cl_raw			= keyp_cl.into_keypair( username=user_cl, password=pswd_cl )

    ls				= subprocess.run(
        [ 'ls', '-l', str( here ) ], stdout=subprocess.PIPE,
    )
    log.info( f"Test directory:\n{ls.stdout.decode( 'UTF-8' )}\n\n" )

    # The licensing.authorized API should now be able to issue this self-signed.crypto-license
    # on-demand, IF it can find a Keypair.  It will not find it by looking for
    # self-signed.crypto-license, b/c self-signed.crypto-keypair is encrypted with different
    # credentials.
    with pytest.raises( licensing.NotRegistered ) as excinfo:
        assert list( licensing.authorized(
            author	= author,
            basename	= str( base_ss ),
            confirm	= False,
            registering	= False,
            acquiring	= False,
        )) == [ (None,None) ]
    # But w/ a Keypair, we can get a self-signed License issued
    auth_str = licensing.into_JSON(
        [
            ( licensing.KeypairPlaintext( k ) if k else None, l )
            for k,l in licensing.authorized(
                author		= author,
                basename	= str( base_ss ),
                confirm		= False,
                registering	= False,
                acquiring	= False,
                keypairs	= [ keyp_cl_raw ],
            )
        ], indent=4, default=str,
    )
    print( auth_str )
    assert auth_str == """\
[
    [
        {
            "sk":"u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7t9WcViPdQKdKpNWjKsZF07P5Xa6uTCK+JUdt1qSG9zgg==",
            "vk":"fVnFYj3UCnSqTVoyrGRdOz+V2urkwiviVHbdakhvc4I="
        },
        {
            "license":{
                "author":{
                    "name":"fVnFYj3UCnSqTVoyrGRdOz+V2urkwiviVHbdakhvc4I=",
                    "pubkey":"fVnFYj3UCnSqTVoyrGRdOz+V2urkwiviVHbdakhvc4I="
                },
                "dependencies":[
                    {
                        "license":{
                            "author":{
                                "domain":"self-signed.com",
                                "name":"Dominion Research & Development Corp.",
                                "product":"Self Signed",
                                "pubkey":"5zTqbCtiV95yNV5HKqBaTEh+a0Y8Ap7TBt8vAbVja1g="
                            },
                            "grant":{
                                "self-signed":{
                                    "some-capability":10
                                }
                            }
                        },
                        "signature":"I/o9+bacFBOwTPgNfduUwdgldMRPVCQPUEN4h3yynqzv4sDyEe37oCslnB7gjM8VBojp3vdZbdlmO7HhxLJQDA=="
                    }
                ]
            },
            "signature":"/OWpu8M4HczWnTXwW5yPchZpdI7B/k7c2GtaiK++itlzm8UY2Gl0DMe2k4HDZY6cg6t1cPi3EgjjtZXrWuLACQ=="
        }
    ],
    [
        {
            "sk":"u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7u7t9WcViPdQKdKpNWjKsZF07P5Xa6uTCK+JUdt1qSG9zgg==",
            "vk":"fVnFYj3UCnSqTVoyrGRdOz+V2urkwiviVHbdakhvc4I="
        },
        null
    ]
]"""

    client			= licensing.Agent(
        name	= "Perry Kundert",
        service	= "client-user",
        pubkey	= keyp_cl['vk'],
    )
    log.info( f"Client: {client}" )

    # We'll be loading an existing Client Agent keypair, so restrict it from registering a new one.
    # We'll look for Licenses issued under the basenames of client.service (an already issued
    # sub-license), and author.service (eg. an author-issued License we can sub-license). If no
    # already issued License is found, tt must try to load the Author's License (using the
    # author.service as the basename), and then attempt to sign and save an instance of it with the
    # client Agent's Keypair.  For this, we need access to an Agent Keypair suitable for signing
    # (access to decrypted private key); so we'll need the credentials.
    machine_id_path		= test.with_suffix( '' ) / "payments_test.machine-id"
    reloader			= reload(
        author		= author,
        client		= client,
        registering	= False,
        reverse_save	= True,
        basename	= None,         # name_cl,  # basename of the License, and the Keypair use to self-sign it
        confirm		= False,
        extra		= [ str( here ) ],
        constraints	= dict(
            machine	= True,
        ),
        machine_id_path	= machine_id_path,
    )

    username			= user_cl
    password			= pswd_cl
    grants			= None
    keypairs,licenses		= [],[]
    try:
        key,val			= next( reloader )
        while True:
            log.info( f"test_grants <-- {key}: {val}" )
            if key == Process.PROMPT:
                if 'username' in val:
                    log.info( f"test_grants --> {username}" )
                    key,val	= reloader.send( username )
                    continue
                elif 'password' in val:
                    log.info( f"test_grants --> {password}" )
                    key,val	= reloader.send( password )
                    continue
                else:
                    log.info( f"test_grants -x- ignoring  {val}" )
            elif key is Process.GRANTS:
                log.warning( f"Grants:  {val}" )
                grants		= val
            elif key is Process.KEYPAIR:
                log.warning( f"Keypair: {val}" )
                keypairs.append( val )
            elif key is Process.LICENSE:
                log.warning( f"License: {val[1]}, w/ Keypair: {licensing.KeypairPlaintext( val[0] )}" )
                keypairs.append( val[0] )
                licenses.append( val[1] )
            key,val		= next( reloader )
    except StopIteration:
        log.info( f"test_grants xxx Done w/ key == {key}, val == {val}" )

    assert str( grants ) == """\
{
    "self-signed":{
        "some-capability":10
    }
}"""
    # TODO: confirm appropriate payment before issuing self-signed license.

    # Now that we've got the self-signed License from Author, self-signed by Client, we can verify
    # the authenticity of the License without access to the private keys; only the public keys.
    ls				= subprocess.run(
        [ 'ls', '-l', str( here ) ], stdout=subprocess.PIPE,
    )
    log.info( f"Test directory (post self-signed issuance):\n{ls.stdout.decode( 'UTF-8' )}\n\n" )

    assert len( licenses ) == 1
    constraints			= licenses[0].verify(
        author_pubkey	= client.pubkey,
        confirm		= False,
        machine_id_path	= machine_id_path,
    )
    log.info( f"Verified self-signed License:\n{constraints}\n\n" )
    assert not constraints

    # Later, we restore from backup and our Machine ID changes...  This will cause a
    # LicenseSigned.verify failure.
    assert len( licenses ) == 1
    with pytest.raises( licensing.LicenseIncompatibility ) as excinfo:
        constraints		= licenses[0].verify(
            author_pubkey	= client.pubkey,
            confirm		= False,
        )
    assert "specifies Machine ID 00010203-0405-4607-8809-0a0b0c0d0e0f; found" in str( excinfo.value )
